#!/usr/bin/perl -w

use strict;
use Data::Dumper;
use Getopt::Std;
use Storable qw(dclone);

use vars qw ($opt_h $opt_i $opt_o $opt_a $opt_b);
&getopts('hi:o:a:b:');

my $usage = <<_EOH_;

## Options ###########################################
## Required:
# -i   anchor file (name.txt)
# -a   plot chromosome ID 1
# -b   plot chromosome ID 2

## Not required:
# -o   output file (default=name.png)

## Others:
# -h   print help
_EOH_
    ;

die $usage if $opt_h;

my $inFile=$opt_i or die $usage;
open(IN,$inFile) or die "Can't open $inFile to read.";
my $plot1=$opt_a or die $usage;
my $plot2=$opt_b or die $usage;
my $outFile;
if($inFile=~/^(.+)\.txt$/){
    $outFile=$1.".".$plot1."vs".$plot2.".png";
}
$outFile=$opt_o if $opt_o;

my ($max1,$max2)=(0,0);
my $plotFile="plot.$$";
open(TMP,">$plotFile") or die "Can't open $plotFile to write.";
while(<IN>){
    chomp;
    my ($chr1,$start1,$end1,$sign1,$chr2,$start2,$end2,$sign2)
	=($1,$2,$3,$4,$5,$6,$7,$8)
	if(/^(-?\d+) (\d+) (\d+) ([+-]) (-?\d+) (\d+) (\d+) ([+-])/);

    if($chr1==$plot1 and $chr2==$plot2){
	$max1=$end1 if $end1>$max1;
	$max2=$end2 if $end2>$max2;
	if($sign1 eq $sign2){
	    print TMP "\n";
	    print TMP "$start1 $start2\n";
	    print TMP "$end1 $end2\n";
	    print TMP "\n";
	} else {
	    print TMP "\n";
	    print TMP "$start1 $end2\n";
	    print TMP "$end1 $start2\n";
	    print TMP "\n";
	}
    }
}
close IN;
close TMP;

my $gnuFile="gnu.$$";
open(GNU,">$gnuFile") or die "Can't open $gnuFile to write.";
print GNU "set terminal png\n";
print GNU "set output \"$outFile\"\n";
print GNU "set xrange [0:$max1]\n";
print GNU "set yrange [0:$max2]\n";
print GNU "plot \"$plotFile\" with lines ti \"\"\n";
close GNU;
`gnuplot $gnuFile`;
`rm $plotFile`;
`rm $gnuFile`;

