// *************************************************************************
//  anc-graph.cpp
// *************************************************************************

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "osfinder.h"

// *************************************************************************
// sort anchor set according to start[0].
void sort_anc_set( std::vector<block>& anc_set, int num_org ){
  std::map< std::vector<int>, std::vector<block> > chrom_hash;
  std::vector< std::vector<int> > chrom_list;

  for( int i=0 ; i<(int)anc_set.size() ; ++i ){
    std::vector<int> chrom_id;
    for( int j=0 ; j<num_org ; ++j ){
      chrom_id.push_back( anc_set[i].chrom_id[j] );
    }
    std::map< std::vector<int>, std::vector<block> >::iterator p;
    p = chrom_hash.find( chrom_id );
    if( p != chrom_hash.end() ){
      p->second.push_back( anc_set[i] );
    } else {
      std::vector<block> tmp_block_vec;
      tmp_block_vec.push_back( anc_set[i] );
      chrom_hash.insert( make_pair(chrom_id, tmp_block_vec ) );
      chrom_list.push_back( chrom_id );
    }
  }

  anc_set.clear();
  for( int i=0 ; i<(int)chrom_list.size() ; ++i ){
    std::vector<block> tmp_block_vec;
    std::map< std::vector<int>, std::vector<block> >::iterator p;
    p = chrom_hash.find( chrom_list[i] );
    if( p != chrom_hash.end() ){
      tmp_block_vec = p->second;
    }

    for( int j=0 ; j<(int)tmp_block_vec.size() ; ++j ){
      int min_idx = j;
      int min_val = tmp_block_vec[j].start[0];
      for( int k=j+1 ; k<(int)tmp_block_vec.size() ; ++k ){
	if( tmp_block_vec[k].start[0] < min_val ){
	  min_idx = k;
	  min_val = tmp_block_vec[k].start[0];
	}
      }
      block tmp_block = tmp_block_vec[j];
      tmp_block_vec[j] = tmp_block_vec[min_idx];
      tmp_block_vec[min_idx] = tmp_block;
    }

    for( int j=0 ; j<(int)tmp_block_vec.size() ; ++j ){
      anc_set.push_back( tmp_block_vec[j] );
    }

    tmp_block_vec.clear();
  }
}

// *************************************************************************
// comparing function of class block.
int comp( block b1, block b2 ){
  return b1.start[0] - b2.start[0];
}

// *************************************************************************
// draw sub-neighbor edges.
void draw_sn_edge( std::vector<edge> &edge_set, const std::vector<block>& anc_set, int num_org, bool prog_bar ){
  if(prog_bar){
    std::cout << " drawing edges" << std::endl;
  }

  std::vector<bool> mask_anc( (int)anc_set.size(), false );
  int total_size = anc_set.size();
  int num_remain = 0;
  int old_prog   = 0;
  while( -1 ){
    num_remain = 0;
    for( int i=0 ; i<(int)mask_anc.size() ; ++i ){
      if( !mask_anc[i] ){
	++num_remain;
      }
    }
    if( num_remain == 0 ){
      break;
    }

    std::vector<block> anc_set_work;
    std::vector<int> look_idx;
    bool flg_first = true;
    block first_anc;
    for( int i=0 ; i<(int)anc_set.size() ; ++i ){
      if( !mask_anc[i] ){
	if( flg_first ){
	  flg_first = false;
	  first_anc = anc_set[i];
	}
	if( !flg_first && is_same_chrom( anc_set[i], first_anc, num_org ) ){
	  anc_set_work.push_back( anc_set[i] );
	  look_idx.push_back( i );
	  mask_anc[i] = true;
	}
      }
    }

    /*
    std::cout << "curent chrom IDs :" << std::endl;
    for( int j=0 ; j<num_org ; j++ ){
      if( j > 0 ){
	std::cout << " ";
      }
      std::cout << anc_set_work[0].chrom_id[j];
    }
    std::cout << " " << "( " << anc_set_work.size() << " )";
    std::cout << std::endl;
    */

    std::vector< std::vector<bool> > mask_set;
    std::vector<bool> tmp_mask_idx(1, false);
    mask_set.push_back(tmp_mask_idx);
    for( int i=1 ; i<(int)anc_set_work.size() ; ++i ){
      if(prog_bar)
	block_set_progress( total_size, 
			    (total_size-num_remain+i-1), old_prog );
      std::vector<bool> mask_idx(i, false);
      for( int j=i-1 ; j>=0 ; --j ){
	if( mask_idx[j] ){
	  continue;
	}
	if( is_colinear( anc_set_work[i], anc_set_work[j], num_org ) ){
	  edge tmp_edge;
	  tmp_edge.idx1 = look_idx[j];
	  tmp_edge.idx2 = look_idx[i];
	  for( int k=0 ; k<num_org ; k++ ){
	    tmp_edge.chrom_id[k] = anc_set_work[i].chrom_id[k];
	    tmp_edge.sign[k] = anc_set_work[i].sign[k];
	  }
	  edge_set.push_back( tmp_edge );

	  mask_idx[j] = true;

	  // mask upper anchors previously found.
	  for( int l=0 ; l<(int)mask_set[j].size() ; ++l ){
	    if( mask_set[j][l] ){
	      mask_idx[l] = true;
	    }
	  }
	}
      }
      mask_set.push_back(mask_idx);
    }
  }
  if(prog_bar){
    std::cout << std::endl
	      << "   ... done" << " (" << edge_set.size() << " edges)"
	      << std::endl;
  }
}

// *************************************************************************
// judge if given two indices of anchor set are same chromosome or not.
bool is_same_chrom( block anc1, block anc2, int num_org){
  bool flg_same_chrom = true;
  for( int i=0 ; i<num_org ; i++ ){
    if( anc1.chrom_id[i] != anc2.chrom_id[i] ){
      flg_same_chrom = false;
      break;
    }
  }

  return flg_same_chrom;
}

// *************************************************************************
// judge if given an anchor and an edge are same chromosome or not.
bool is_same_chrom( block a, edge e, int num_org){
  bool flg_same_chrom = true;
  for( int i=0 ; i<num_org ; i++ ){
    if( a.chrom_id[i] != e.chrom_id[i] ){
      flg_same_chrom = false;
      break;
    }
  }

  return flg_same_chrom;
}

// *************************************************************************
// judge if given two anchors are co-linear or not.
bool is_colinear( block anc1, block anc2, int num_org ){
  bool flg_same_direc;
  bool flg_share_space;
  bool flg_overlap;

  flg_same_direc = is_same_direc( anc1, anc2, num_org );
  flg_share_space = is_share_space( anc1, anc2, num_org );
  flg_overlap = is_overlap( anc1, anc2, num_org );

  if( flg_same_direc && flg_share_space && !flg_overlap ){
    return true;
  } else {
    return false;
  }
}

// *************************************************************************
// judge if given two anchors are the same direction or not.
bool is_same_direc( block anc1, block anc2, int num_org ){
  bool flg_all_same = true;
  bool flg_all_differ = true;

  for( int i=0 ; i<num_org ; i++ ){
    if( anc1.sign[i] == anc2.sign[i] ){
      flg_all_differ = false;
    } else {
      flg_all_same = false;
    }
  }

  if( flg_all_same || flg_all_differ ){
    return true;
  } else {
    return false;
  }
}


// *************************************************************************
// judge if given two anchors share space.
bool is_share_space( block anc1, block anc2, int num_org ){
  bool flg_all_large_1 = true;
  bool flg_all_large_2 = true;

  for( int i=0 ; i<num_org ; i++ ){
    if( anc1.end[i] < anc2.start[i] ){
      anc1.sign[i] == '+' ? flg_all_large_1 = false : flg_all_large_2 = false;
    } else if ( anc2.end[i] < anc1.start[i] ){
      anc1.sign[i] == '+' ? flg_all_large_2 = false : flg_all_large_1 = false;
    }
  }

  if( flg_all_large_1 || flg_all_large_2 ){
    return true;
  } else {
    return false;
  }
}


// *************************************************************************
// judge if given two anchors are overlap or not.
bool is_overlap( block anc1, block anc2, int num_org ){
  bool flg_overlap = false;

  for( int i=0 ; i<num_org ; i++ ){
    if( anc1.chrom_id[i] != anc2.chrom_id[i] ){ continue; }
    if( anc1.end[i] < anc2.start[i] || anc2.end[i] < anc1.start[i] ){
    } else {
      flg_overlap = true;
    }
  }

  return flg_overlap;
}


// *************************************************************************
// judge if given two anchors are overlap on gen_idx genome or not.
bool is_overlap( block anc1, block anc2, int gen_idx, int num_org ){
  bool flg_overlap = false;

  int i = gen_idx;
  if( anc1.end[i] < anc2.start[i] || anc2.end[i] < anc1.start[i] ){
  } else {
    flg_overlap = true;
  }

  return flg_overlap;
}


// *************************************************************************
// judge if given two anchors are overlap among all genomes.
bool is_all_overlap( block anc1, block anc2, int num_org ){
  bool flg_all_overlap = true;

  for( int i=0 ; i<num_org ; i++ ){
    if( !is_overlap( anc1, anc2, i, num_org ) ){
      flg_all_overlap = false;
    }
  }

  return flg_all_overlap;
}

