// *************************************************************************
//  chaining.cpp
// *************************************************************************

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "osfinder.h"

// *************************************************************************
// join two sub-neighbor anchors whose distance is less than a % of chromosome length.
void chaining_all( std::vector<block>& clst_set, const std::vector<block>& anc_set, const std::vector<edge>& sn_edge_set, int num_org ){
  std::cout << " chaining anchors" << std::endl;

  int total_size = (int)anc_set.size(); 
  int old_prog   = 0;
  bool flg_prog  = false;
  std::vector<bool> mask_anc( total_size, false );
  while( -1 ){
    int num_remain = 0;
    for( int i=0 ; i<(int)mask_anc.size() ; ++i ){
      if( !mask_anc[i] ){
	++num_remain;
      }
    }
    if( num_remain == 0 ){
      break;
    }

    if( (total_size - num_remain) > 0){
      if(!flg_prog){
	std::cout << " 0 ";
	flg_prog = true;
      }
      block_set_progress( total_size, (total_size-num_remain), old_prog );
    }

    std::vector<block> anc_set_work;
    std::vector<int> look_idx;
    std::map<int, int> rev_look_idx;
    std::vector<edge> edge_set_work;
    std::vector<block> clst_set_work;
    bool flg_first = true;
    block first_anc;
    for( int i=0 ; i<(int)anc_set.size() ; ++i ){
      if( !mask_anc[i] ){
	if( flg_first ){
	  flg_first = false;
	  first_anc = anc_set[i];
	}
	if( !flg_first && is_same_chrom( anc_set[i], first_anc, num_org ) ){
	  anc_set_work.push_back( anc_set[i] );
	  rev_look_idx.insert( std::make_pair( i, (int)look_idx.size() ) );
	  look_idx.push_back( i );
	  mask_anc[i] = true;
	}
      }
    }
    bool flg_first_same = false;
    for( int i=0 ; i<(int)sn_edge_set.size() ; ++i ){
      if( is_same_chrom( first_anc, sn_edge_set[i], num_org ) ){
	flg_first_same = true;
	edge tmp_edge = sn_edge_set[i];
	std::map<int, int>::iterator p;
	p = rev_look_idx.find(sn_edge_set[i].idx1);
	tmp_edge.idx1 = p->second;
	p = rev_look_idx.find(sn_edge_set[i].idx2);
	tmp_edge.idx2 = p->second; 
	if( tmp_edge.idx1 < 0 || tmp_edge.idx2 < 0 ){
	  std::cerr << tmp_edge.idx1 << " " << tmp_edge.idx2 << std::endl;
	  exit(1);
	}
	edge_set_work.push_back( tmp_edge );
      } else if( flg_first_same  ){
	break;
      }
    }

    chaining( clst_set_work, anc_set_work, edge_set_work, num_org );

    for( int i=0 ; i<(int)clst_set_work.size() ; ++i ){
      block tmp_clst = clst_set_work[i];
      for( int j=0 ; j<(int)tmp_clst.anc_idx_set.size() ; ++j ){
	tmp_clst.anc_idx_set[j] = look_idx[tmp_clst.anc_idx_set[j]];
      }
      clst_set.push_back( tmp_clst );
    }

    anc_set_work.clear();
    edge_set_work.clear();
    look_idx.clear();
    rev_look_idx.clear();
    clst_set_work.clear();
  }
  if(flg_prog)
    std::cout << std::endl;
  std::cout << "   ... done" << std::endl;
}

// *************************************************************************
// join two sub-neighbor anchors.
void chaining( std::vector<block>& clst_set, const std::vector<block>& anc_set, const std::vector<edge>& sn_edge_set, int num_org ){
  dp_cell* dp_tbl;

  try {
    dp_tbl = new dp_cell [anc_set.size()];
    if( dp_tbl == NULL ) throw dp_tbl;
  } catch (...){
    std::cerr << "Memory Error when chaining anchors!!" << std::endl;
    exit(1);
  }

  bool flg_extract = true;
  std::vector<bool> mask_edge( (int)sn_edge_set.size(), false );
  while( flg_extract ){
    init_chtbl( dp_tbl, anc_set );
    rfn_chtbl( dp_tbl, anc_set, sn_edge_set, mask_edge, num_org );

    flg_extract = false;
    for( int i=0 ; i<(int)anc_set.size() ; i++ ){
      if( dp_tbl[i].sin_scr < dp_tbl[i].chn_scr ){
	flg_extract = true;
	break;
      }
    }
    if( !flg_extract ){
      break;
    }

    block tmp_cls;
    tmp_cls = trb_chtbl( anc_set, dp_tbl, num_org );
    clst_set.push_back( tmp_cls );

    for( int i=0 ; i<(int)anc_set.size() ; ++i ){
      if( is_all_overlap( tmp_cls, anc_set[i], num_org ) ){
	for( int j=0 ; j<(int)sn_edge_set.size() ; ++j ){
	  if( sn_edge_set[j].idx1 == i || sn_edge_set[j].idx2 == i ){
	    mask_edge[j] = true;
	  }
	}
      }
    }
  }
  mask_edge.clear();
  delete [] dp_tbl;
}

// *************************************************************************
// initialize chain dp table.
void init_chtbl( dp_cell* dp_tbl, const std::vector<block>& anc_set ){
  for( int i=0 ; i<(int)anc_set.size() ; i++ ){
    dp_tbl[i].sin_scr = anc_set[i].score;
    dp_tbl[i].chn_scr = dp_tbl[i].sin_scr;
    dp_tbl[i].trb_idx = -1;
    dp_tbl[i].anc_idx_set.clear();
    dp_tbl[i].anc_idx_set.push_back( i );
  }
}

// *************************************************************************
// refine chain dp table.x
void rfn_chtbl( dp_cell* dp_tbl, const std::vector<block>& anc_set, const std::vector<edge>& sn_edge_set, const std::vector<bool>& mask_edge, int  num_org ){
  for( int i=0 ; i<(int)sn_edge_set.size() ; i++ ){
    if( mask_edge[i] ){
      continue;
    }

    int idx1 = sn_edge_set[i].idx1;
    int idx2 = sn_edge_set[i].idx2;
    assert( is_same_chrom( anc_set[idx1], anc_set[idx2], num_org ) );
    if( anc_set[idx1].start[0] > anc_set[idx2].start[0] ){
      int tmp = idx1;
      idx1 = idx2;
      idx2 = tmp;
    }

    double new_scr = dp_tbl[idx1].chn_scr + sn_edge_set[i].score + dp_tbl[idx2].sin_scr;
    double diff_scr = sn_edge_set[i].score + dp_tbl[idx2].sin_scr;
    if( diff_scr > 0 && new_scr >= dp_tbl[idx2].chn_scr ){
      dp_tbl[idx2].chn_scr = new_scr;
      dp_tbl[idx2].trb_idx = idx1;
      dp_tbl[idx2].anc_idx_set.clear();
      for( int i=0 ; i<(int)dp_tbl[idx1].anc_idx_set.size() ; i++ ){
	int anc_idx = dp_tbl[idx1].anc_idx_set[i];
	dp_tbl[idx2].anc_idx_set.push_back( anc_idx );
      }
      dp_tbl[idx2].anc_idx_set.push_back( idx2 );
    }
  }
}

// *************************************************************************
// traceback chain dp table.
block trb_chtbl( const std::vector<block>& anc_set, dp_cell* dp_tbl, int num_org ){
  bool def_max_scr = false;
  double max_scr = -1;
  int max_idx = -1;
  for( int i=0 ; i<(int)anc_set.size() ; i++ ){
    if( dp_tbl[i].sin_scr < dp_tbl[i].chn_scr ){
      if( !def_max_scr || dp_tbl[i].chn_scr > max_scr ){
	max_scr = dp_tbl[i].chn_scr;
	max_idx = i;
	def_max_scr = true;
      }
    }
  }

  assert( max_idx > -1 );
  int cur_idx = max_idx;
  block tmp_cls;
  tmp_cls.score = max_scr;
  tmp_cls.anc_idx_set.push_back( cur_idx );
  tmp_cls.chn_scr = max_scr;
  for( int i=0 ; i<num_org ; i++ ){
    tmp_cls.chrom_id[i] = anc_set[cur_idx].chrom_id[i];
    tmp_cls.start[i] = anc_set[cur_idx].start[i];
    tmp_cls.end[i]   = anc_set[cur_idx].end[i];
    tmp_cls.sign[i]  = anc_set[cur_idx].sign[i];
  }
  
  for( int i=0 ; i<(int)dp_tbl[cur_idx].anc_idx_set.size()-1 ; i++ ){
    int anc_idx = dp_tbl[cur_idx].anc_idx_set[i];
    tmp_cls.anc_idx_set.push_back( anc_idx );
    for( int j=0 ; j<num_org ; j++ ){
      if( anc_set[anc_idx].start[j] < tmp_cls.start[j] ){
	tmp_cls.start[j] = anc_set[anc_idx].start[j];
      }
      if( anc_set[anc_idx].end[j] > tmp_cls.end[j] ){
	tmp_cls.end[j] = anc_set[anc_idx].end[j];
      }
    }
  }
  
  assert( (int)tmp_cls.anc_idx_set.size() > 1 );
  return tmp_cls;
}

// *************************************************************************
// make chained anchor set.
void mk_chained_anc_set( std::vector<block>& chained_anc_set, const std::vector<block>& clst_set, const std::vector<block>& anc_set ){
  for( int i=0 ; i<(int)clst_set.size() ; i++ ){
    for( int j=0 ; j<(int)clst_set[i].anc_idx_set.size() ; j++ ){
      int anc_idx = clst_set[i].anc_idx_set[j];
      chained_anc_set.push_back( anc_set[anc_idx] );
    }
  }
}

// *************************************************************************
// make the set of anchors which are included in syntenic blocks.
void mk_anc_in_sb_set( std::vector<block>& anc_in_sb_set, const std::vector<block>& sb_set, const std::vector<block>& clst_set, const std::vector<block>& anc_set ){
  for( int i=0 ; i<(int)sb_set.size() ; ++i ){
    for( int j=0 ; j<(int)sb_set[i].anc_idx_set.size() ; ++j ){
      int clst_idx = sb_set[i].anc_idx_set[j];
      for( int k=0 ; k<(int)clst_set[clst_idx].anc_idx_set.size() ; ++k ){
	int anc_idx = clst_set[clst_idx].anc_idx_set[k];
	anc_in_sb_set.push_back( anc_set[anc_idx] );
      }
    }
  }
}

// *************************************************************************
// join two sub-neighbor SHRs.
void chaining_shr( std::vector<block>& sb_set, const std::vector<block>& clst_set, int sb_size, int num_org ){
  std::vector<edge> shr_edge_set;
  for( int i=0 ; i<(int)clst_set.size() ; ++i ){
    sb_set.push_back( clst_set[i] );
    sb_set[i].anc_idx_set.clear();
    sb_set[i].anc_idx_set.push_back( i );
    sb_set[i].cal_len( num_org );
    // sb_set[i].score = sb_set[i].len;
  }
  int old_sb_set_size = 0;
  int sb_set_size = sb_set.size();

  bool flgFirstChain=false;
  while( sb_set_size != old_sb_set_size ){
    shr_edge_set.clear();
    old_sb_set_size = sb_set_size;
    sort_anc_set( sb_set, num_org );
    draw_sn_edge( shr_edge_set, sb_set, num_org );

    for( int i=0 ; i<(int)shr_edge_set.size() ; ++i ){
      shr_edge_set[i].cal_len( sb_set, num_org );
    }

    sort_edge_short2long( shr_edge_set );
    std::vector<bool> mask_sb( (int)sb_set.size(), false );
    std::vector<bool> mask_edge( (int)shr_edge_set.size(), false );
    if(!flgFirstChain){
      std::multimap<double,int> scr2idx;
      calIdxOrder(scr2idx,sb_set);
      for(std::multimap<double,int>::reverse_iterator mmi=scr2idx.rbegin();mmi!=scr2idx.rend();++mmi){
	int sbIdx=mmi->second;
	if(mask_sb[sbIdx]){continue;}
	if( is_large_sb( sb_set[sbIdx], sb_size, num_org ) ){
	  remove_overlap(sb_set,sbIdx,num_org,shr_edge_set,mask_sb,mask_edge);
	} else {
	  break;
	}
      }
    }

    for( int i=0 ; i<(int)shr_edge_set.size() ; ++i ){
      if( mask_edge[i] ){ continue; }
      int idx1 = shr_edge_set[i].idx1;
      int idx2 = shr_edge_set[i].idx2;
      merge_sb(sb_set, idx1, idx2, num_org, shr_edge_set, mask_sb, mask_edge);
      if( is_large_sb( sb_set[idx1], sb_size, num_org ) ){
	remove_overlap(sb_set,idx1,num_org,shr_edge_set,mask_sb,mask_edge);
      }
    }
    remove_mask_block( sb_set, mask_sb );
    sb_set_size = sb_set.size();
    flgFirstChain=false;
  }
}

// *************************************************************************
// sort sub-neighbor SHRs according to length of SHRs.
void sort_edge_short2long( std::vector<edge>& shr_edge_set ){
  for( int i=0 ; i<(int)shr_edge_set.size() ; ++i ){
    int min_len = shr_edge_set[i].len;
    int min_idx = i;
    for( int j=i+1 ; j<(int)shr_edge_set.size() ; ++j ){
      if( shr_edge_set[j].len < min_len ){
	min_len = shr_edge_set[j].len;
	min_idx = j;
      }
    }
    edge tmp_edge = shr_edge_set[min_idx];
    shr_edge_set[min_idx] = shr_edge_set[i];
    shr_edge_set[i] = tmp_edge;
  }
}

// *************************************************************************
// sort anchors according to anchor length
void sort_anchor_short2long( std::vector<block>& anchor_set ){
  for( int i=0 ; i<(int)anchor_set.size() ; ++i ){
    int min_len = anchor_set[i].len;
    int min_idx = i;
    for( int j=i+1 ; j<(int)anchor_set.size() ; ++j ){
      if( anchor_set[j].len < min_len ){
	min_len = anchor_set[j].len;
	min_idx = j;
      }
    }
    block tmp_anc = anchor_set[min_idx];
    anchor_set[min_idx] = anchor_set[i];
    anchor_set[i] = tmp_anc;
  }
}

// *************************************************************************
// Merge two syntenic blocks.
void merge_sb( std::vector<block>& sb_set, int idx1, int idx2, int num_org, std::vector<edge>& shr_edge_set, std::vector<bool>& mask_sb, std::vector<bool>& mask_edge ){
  block merged_sb;
  block sb1 = sb_set[idx1];
  block sb2 = sb_set[idx2];
  for( int i=0 ; i<num_org ; ++i ){
    assert( sb1.chrom_id[i] == sb2.chrom_id[i] );
    merged_sb.chrom_id[i] = sb1.chrom_id[i];
    merged_sb.sign[i] = sb1.sign[i];
    merged_sb.start[i] = min( sb1.start[i], sb2.start[i] );
    merged_sb.end[i] = max( sb1.end[i], sb2.end[i] );
  }
  for( int i=0 ; i<(int)sb_set[idx1].anc_idx_set.size() ; ++i ){
    int anc_idx = sb_set[idx1].anc_idx_set[i];
    merged_sb.anc_idx_set.push_back( anc_idx );
  }

  sb_set[idx1] = merged_sb;
  for( int i=0 ; i<(int)sb_set[idx2].anc_idx_set.size() ; ++i ){
    int anc_idx = sb_set[idx2].anc_idx_set[i];
    sb_set[idx1].anc_idx_set.push_back( anc_idx );
    sb_set[idx1].score += sb_set[idx2].score;
  }

  delete_sb( sb_set, idx1, idx2, shr_edge_set, mask_sb, mask_edge );
}

// *************************************************************************
// Judge if the size of a syntenic block is larger than minimum syntenic 
// block size, or not.
bool is_large_sb( block sb, int sb_size, int num_org ){
  bool flg_large_sb = true;
  for( int i=0 ; i<num_org ; ++i ){
    int len = sb.end[i] - sb.start[i] + 1;
    if( len < sb_size ){
      flg_large_sb = false;
      break;
    }
  }
  return flg_large_sb;
}

// *************************************************************************
// Delete a syntenic block from syntenic block set.
void delete_sb( std::vector<block>& sb_set, int mer_idx, int del_idx, std::vector<edge>& shr_edge_set, std::vector<bool>& mask_sb, std::vector<bool>& mask_edge ){
  mask_sb[del_idx] = true;

  for( int i=0 ; i<(int)shr_edge_set.size() ; ++i ){
    if( mask_edge[i] ){ continue; }

    if( shr_edge_set[i].idx1 == mer_idx ){
      mask_edge[i] = true;
    } else if( shr_edge_set[i].idx2 == del_idx ){
      mask_edge[i] = true;
    } else if( shr_edge_set[i].idx1 == del_idx ){
      shr_edge_set[i].idx1 = mer_idx;
    }
  }
} 

// *************************************************************************
// Remove syntenic blocks overlapping with the merged syntenic block
// and sb_DAG_edges whose start or end node is equal to the removed syntenic 
// blocks.
void remove_overlap( std::vector<block>& sb_set, int mer_idx, int num_org, std::vector<edge>& shr_edge_set, std::vector<bool>& mask_sb, std::vector<bool>& mask_edge ){
  for( int i=0 ; i<(int)sb_set.size() ; ++i ){
    if( mask_sb[i] ){ continue; }
    if( i == mer_idx ){ continue; }
    if( is_overlap( sb_set[i], sb_set[mer_idx], num_org ) ){
      mask_sb[i] = true;
      
      for( int j=0 ; j<(int)shr_edge_set.size() ; ++j ){
	if( mask_edge[j] ){ continue; }
	if( shr_edge_set[j].idx1 == i || shr_edge_set[j].idx2 == i ){
	  mask_edge[j] = true;
	}
      }
    }
  }

  for( int i=0 ; i<(int)shr_edge_set.size() ; ++i ){
    if( mask_edge[i] ){ continue; }
    int idx1 = shr_edge_set[i].idx1;
    int idx2 = shr_edge_set[i].idx2;
    block edge_region;
    for( int j=0 ; j<num_org ; ++j ){
      assert( sb_set[idx1].chrom_id[j] == sb_set[idx2].chrom_id[j] );
      edge_region.chrom_id[j] = sb_set[idx1].chrom_id[j];
      edge_region.sign[j] = sb_set[idx1].sign[j];
      edge_region.start[j] = min(sb_set[idx1].end[j], sb_set[idx2].end[j])+1;
      edge_region.end[j] = max(sb_set[idx1].start[j], sb_set[idx2].start[j])-1;
    }
    if( is_overlap( edge_region, sb_set[mer_idx], num_org ) ){
      mask_edge[i] = true;
    }
  }
}

// *************************************************************************
// Remove (syntenic) masked blocks.
void remove_mask_block( std::vector<block>& sb_set, const std::vector<bool>& mask_sb ){
  int num_rem = 0;
  for( int i=0 ; i<(int)sb_set.size() ; ++i ){
    if( mask_sb[i+num_rem] ){
      std::vector<block>::iterator p = sb_set.begin() + i;
      sb_set.erase( p );
      --i;
      ++num_rem;
    }
  }
}

// *************************************************************************
// Remove (syntenic) blocks whose size is less than the specified minimum 
// threshold.
void size_filter( std::vector<block>& sb_set, int sb_size, int num_org ){
  for( int i=0 ; i<(int)sb_set.size() ; ++i ){
    if( !is_large_sb( sb_set[i], sb_size, num_org ) ){
      std::vector<block>::iterator p = sb_set.begin() + i;
      sb_set.erase( p );
      --i;
    }
  }
}

// *************************************************************************
// Remove (syntenic) blocks whose size is less than the specified minimum 
// threshold.
void calIdxOrder(std::multimap<double,int>& scr2idx,std::vector<block>& sb_set){
  for(int i=0;i<(int)sb_set.size();++i){
    scr2idx.insert(std::pair<double,int>(sb_set[i].score,i));
  }
}
