// *************************************************************************
//  main.cpp
// *************************************************************************

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "osfinder.h"

// *************************************************************************
// main function.

int main( int argc, char* argv[] ){
  input in;

  in = option_disposal( argc, argv );
  in.print();

  std::cout << std::endl;
  std::cout << "** reading anchor file *************************" << std::endl;
  int num_org = in.get_num_org();
  std::vector<block> anc_set;
  std::cout << " reading anchor file" << std::endl;
  get_anc( anc_set, in.get_ancfile(), num_org );
  std::cout << "   ... done" 
	    << " (" << anc_set.size() << " anchors)" 
	    << std::endl;
  if( !in.get_flagS() ){
    std::cout << " sorting anchors" << std::endl;
    sort_anc_set( anc_set, num_org );
    std::cout << "   ... done" << std::endl;
  }
  for( int i=0 ; i<(int)anc_set.size() ; ++i ){
    anc_set[i].cal_len( num_org );
  }
  std::cout << "************************************************" << std::endl;

  std::cout << "** constructing anchor graph *******************" << std::endl;
  std::vector<edge> sn_edge_set;
  draw_sn_edge( sn_edge_set, anc_set, num_org, true );
  for( int i=0 ; i<(int)sn_edge_set.size() ; ++i ){
    sn_edge_set[i].cal_len( anc_set, num_org );
  }
  std::cout << "************************************************" << std::endl;

  double Ps_mg, Ps_gm, Pr_mg, Pr_gm;

  std::cout << "** optimizing paramerters **********************" << std::endl;
  if( in.get_flagL() ){
    // local likelihood maximization
    std::cout << " performing local maximization algorithm" << std::endl;
    initialize(Ps_mg, Ps_gm, Pr_mg, Pr_gm, anc_set, sn_edge_set, num_org);
    std::cout << std::setprecision( 10 ); 
    int num_rep = 0;
    for( int i=0 ; i<LEARNING_REPEAT ; i++ ){
      /*
      std::cout << " repeat number = " << i << std::endl;
      std::cout << " Pan in SHR    = " << Ps_mg << std::endl;
      std::cout << " Pna in SHR    = " << Ps_gm << std::endl;
      std::cout << " Pan in NSHR   = " << Pr_mg << std::endl;
      std::cout << " Pna in NSHR   = " << Pr_gm << std::endl;
      */
      double old_Ps_mg = Ps_mg;
      double old_Ps_gm = Ps_gm;
      double old_Pr_mg = Pr_mg;
      double old_Pr_gm = Pr_gm;
      
      learn( sn_edge_set, anc_set, Ps_mg, Ps_gm, Pr_mg, Pr_gm, num_org );
      
      if( old_Ps_mg == Ps_mg && old_Ps_gm == Ps_gm && old_Pr_mg == Pr_mg && old_Pr_gm == Pr_gm){
	break;
      }
      num_rep = i;
    }
    std::cout << "   ... done" 
	      << " (" << num_rep << " iterations)" 
	      << std::endl;
  } else {
    // global likelihood maximization
    std::cout << " performing global maximization algorithm" << std::endl;
    learn_anchor_model( anc_set, Ps_mg, Pr_mg, num_org );
    learn_edge_model( sn_edge_set, Ps_gm, Pr_gm, num_org );
    std::cout << "   ... done" << std::endl;
  }
  std::cout << "************************************************" << std::endl;


  double Ps_mm = 1.0 - Ps_mg;
  double Ps_gg = 1.0 - Ps_gm;
  double Pr_mm = 1.0 - Pr_mg;
  double Pr_gg = 1.0 - Pr_gm;
  std::vector<block> clst_set;
  /*
  std::cout << "** chaining paramerters **********************" << std::endl;
  std::cout << " Pan in SHR    = " << Ps_mg << std::endl;
  std::cout << " Paa in SHR    = " << Ps_mm << std::endl;
  std::cout << " Pna in SHR    = " << Ps_gm << std::endl;
  std::cout << " Pnn in SHR    = " << Ps_gg << std::endl;
  std::cout << " Pan in NSHR   = " << Pr_mg << std::endl;
  std::cout << " Paa in NSHR   = " << Pr_mm << std::endl;
  std::cout << " Pna in NSHR   = " << Pr_gm << std::endl;
  std::cout << " Pnn in NSHR   = " << Pr_gg << std::endl;
  std::cout << "**********************************************" << std::endl;
  */

  std::cout << "** extracting chains ***************************" << std::endl;
  std::cout << " assigning scores to anchors and edges" << std::endl;
  double ancExtend_scr = log( Ps_mm / Pr_mm );
  double ancOpen_scr   = log( Ps_mg / Pr_mg );
  double edgeExtend_scr = log( Ps_gg / Pr_gg );
  double edgeOpen_scr   = log( Ps_gm / Pr_gm );
  assignAncScr( anc_set, ancExtend_scr, ancOpen_scr, num_org );
  assignEdgeScr( sn_edge_set, edgeExtend_scr, edgeOpen_scr, num_org );
  std::cout << "   ... done" << std::endl;

  // std::cout << " chaining anchors" << std::endl;
  std::vector<block> chained_anc_set;
  chaining_all( clst_set, anc_set, sn_edge_set, num_org );
  sort_anc_set( clst_set, num_org );
  mk_chained_anc_set( chained_anc_set, clst_set, anc_set );
  if( in.get_flagD() ){
    prt_anc( chained_anc_set, in.get_inchainfile(), num_org );
  }
  // std::cout << "   ... done" << std::endl;
    
  std::cout << " printing the coordinate spans for chains" << std::endl;
  // prt_anc( clst_set, num_org );
  prt_anc( clst_set, in.get_chainfile(), num_org );
  for( int i=0 ; i<(int)clst_set.size() ; ++i ){
    clst_set[i].cal_len( num_org );
  }
  if( in.get_flagD() ){
    prt_chain_detail( clst_set, anc_set, in.get_chaindetailfile(), num_org );
  }
  std::cout << "   ... done" 
	    << " (" << clst_set.size() << " chains)"
	    << std::endl;
  std::cout << "************************************************" << std::endl;

  // from chain to orthologous segments
  std::cout << "** calculating orthologous segments ************" << std::endl;
  std::cout << " constructing chain graph" << std::endl;
  int sb_size = in.get_sb_size();
  std::vector<block> sb_set;
  std::vector<block> anc_in_sb_set;
  std::vector<block> chained_shr_set;
  chaining_shr( sb_set, clst_set, sb_size, num_org );
  size_filter( sb_set, sb_size, num_org );
  mk_anc_in_sb_set( anc_in_sb_set, sb_set, clst_set, anc_set );
  if( in.get_flagD() ){
    prt_anc( anc_in_sb_set, in.get_inosfile(), num_org );
  }
  mk_chained_anc_set( chained_shr_set, sb_set, clst_set );
  // prt_anc( sb_set, num_org );
  std::cout << "   ... done" << std::endl;
    
  std::cout << " printing orthologous segments" << std::endl;
  //  prt_anc( sb_set, num_org );
  prt_anc( sb_set, in.get_osfile(), num_org );
  for( int i=0 ; i<(int)sb_set.size() ; ++i ){
    sb_set[i].cal_len( num_org );
  }
  if( in.get_flagD() ){
    prt_os_detail( sb_set, clst_set, in.get_osdetailfile(), num_org );
  }
  std::cout << "   ... done"
	    << " (" << sb_set.size() << " orthologous segments)"
	    << std::endl;
  std::cout << "************************************************" << std::endl;
    
  return 0;
}
