// *************************************************************************
//  print.cpp
// *************************************************************************

#ifdef HAVE_CONFIG_H
#include "config.h"
#endif

#include "osfinder.h"

// *************************************************************************
// print options.
void input::print(){
  std::ofstream fout( optfile );
  fout << " anchor file               -> " << ancfile << std::endl;
  fout << " chain file                -> " << chainfile << std::endl;
  fout << " orthologous segment file  -> " << osfile << std::endl;
  fout << " number of genomes         -> " << num_org << std::endl;
  fout << " minimum segment length    -> " << sb_size << std::endl;
  fout << " learning method           -> ";
  if( flagL ) fout << "local maximization method";
  else fout << "global maximization method";
  fout << std::endl;
  fout << " assumption                -> ";
  fout << "input anchors are ";
  if( !flagS ) fout << "not ";
  fout << "sorted" << std::endl;
  fout << " print detailed result     -> ";
  if( flagD ) fout << "yes";
  else fout << "no";
  fout << std::endl;
}


// *************************************************************************
// print anchors.
void prt_anc( std::vector<block>& anc_set, int num_org ){
  for( int i=0 ; i<(int)anc_set.size() ; i++ ){
    anc_set[i].show( num_org );
  }
}

// *************************************************************************
// print function which is belong to anchor class.
void block::show( int num_org ){
  for( int i=0 ; i<num_org ; i++ ){
    if( i ){
      std::cout << ' ';
    }
    std::cout << chrom_id[i] << ' '<< start[i] << ' ' << end[i] << ' ' << sign[i];
  }
  std::cout << std::endl;
}

// *************************************************************************
// print anchors to fout.
void prt_anc( const std::vector<block>& anc_set, char* blockfile, int num_org ){
  std::ofstream fout( blockfile );

  for( int i=0 ; i<(int)anc_set.size() ; i++ ){
    for( int j=0 ; j<num_org ; j++ ){
      if( j ){
	fout << ' ';
      }
      fout << anc_set[i].chrom_id[j] << ' ' << anc_set[i].start[j] << ' ' << anc_set[i].end[j] << ' ' << anc_set[i].sign[j];
    }
    fout << std::endl;
  }
  fout.close();
}

// *************************************************************************
// print anchors to fout in detail.
void prt_chain_detail( const std::vector<block>& cls_set, const std::vector<block>& anc_set, char* detailfile, int num_org ){
  std::ofstream fout( detailfile );

  for( int i=0 ; i<(int)cls_set.size() ; i++ ){
    fout << "chain> ";
    for( int j=0 ; j<num_org ; j++ ){
      if( j ){
	fout << ' ';
      }
      fout << cls_set[i].chrom_id[j] << ' ' << cls_set[i].start[j] << ' ' << cls_set[i].end[j] << ' ' << cls_set[i].sign[j];
    }
    fout << std::endl;

    fout << "score> " << cls_set[i].score << std::endl;
    fout << "num_anchor> " << (int)cls_set[i].anc_idx_set.size() << std::endl;

    for( int j=0 ; j<(int)cls_set[i].anc_idx_set.size() ; ++j ){
      fout << "anchor> ";
      for( int k=0 ; k<num_org ; ++k ){
	if( k ){
	  fout << ' ';
	}
	int anc_idx = cls_set[i].anc_idx_set[j];
	fout << anc_set[anc_idx].chrom_id[k] << ' ' << anc_set[anc_idx].start[k] << ' ' << anc_set[anc_idx].end[k] << ' ' << anc_set[anc_idx].sign[k];
      }
      fout << std::endl;
    }
    fout << std::endl;
  }
  fout.close();
}

// *************************************************************************
// print chains to fout in detail.
void prt_os_detail( const std::vector<block>& cls_set, const std::vector<block>& anc_set, char* detailfile, int num_org ){
  std::ofstream fout( detailfile );

  for( int i=0 ; i<(int)cls_set.size() ; i++ ){
    fout << "orthologous segment> ";
    for( int j=0 ; j<num_org ; j++ ){
      if( j ){
	fout << ' ';
      }
      fout << cls_set[i].chrom_id[j] << ' ' << cls_set[i].start[j] << ' ' << cls_set[i].end[j] << ' ' << cls_set[i].sign[j];
    }
    fout << std::endl;

    fout << "num_chain> " << (int)cls_set[i].anc_idx_set.size() << std::endl;

    for( int j=0 ; j<(int)cls_set[i].anc_idx_set.size() ; ++j ){
      fout << "chain> ";
      for( int k=0 ; k<num_org ; ++k ){
	if( k ){
	  fout << ' ';
	}
	int anc_idx = cls_set[i].anc_idx_set[j];
	fout << anc_set[anc_idx].chrom_id[k] << ' ' << anc_set[anc_idx].start[k] << ' ' << anc_set[anc_idx].end[k] << ' ' << anc_set[anc_idx].sign[k];
      }
      fout << std::endl;
    }
    fout << std::endl;
  }
  fout.close();
}

// *************************************************************************
// print anchor length.
void prt_anc_len( const std::vector<block>& anc_set, char* anc_len_file ){
  std::ofstream fout( anc_len_file );
  for( int i=0 ; i<(int)anc_set.size() ; ++i ){
    fout << anc_set[i].len << std::endl;
  }
  fout.close();
}

// *************************************************************************
// print edge length.
void prt_edge_len( const std::vector<edge>& edge_set, char* edge_len_file ){
  std::ofstream fout( edge_len_file );
  for( int i=0 ; i<(int)edge_set.size() ; ++i ){
    fout << edge_set[i].len << std::endl;
  }
  fout.close();
}

// *************************************************************************
// print pregress of some operration to block set.
void block_set_progress( int total_size, int idx, int& old_prog ){
  if( idx == 0 ){
    std::cout << " 0 ";
    old_prog = 0;
  } else {
    int new_prog = idx * 100 / total_size;
    if( old_prog < new_prog ){
      int old_i = old_prog / 10;
      int old_j = old_prog % 10;
      int new_i = new_prog / 10;
      int new_j = new_prog %10;
      for( int i=old_i+1 ; i<=new_i ; i++ ){
	std::cout << std::endl;
	std::cout << (i*10) << " ";
      }
      for( int j=old_j+1 ; j<=new_j ; j++ ){
	std::cout << "*";
      }
    }
    old_prog = new_prog;
  }
  std::cout.flush();
}
